const webpack = require('webpack')
const chalk = require('chalk')
const formatWebpackMessages = require('./formatWebpackMessages')
const clearConsole = require('./clearConsole')
const { spawn } = require('child_process')
const path = require('path')

const isInteractive = process.stdout.isTTY

function createCompiler(webpack, config, appName) {
    let app
    let spawnedApp = false
    let compiler
    try {
        compiler = webpack(config)
    } catch (err) {
        console.log(chalk.red('Failed to compile.'))
        console.log()
        console.log(err.message || err)
        console.log()
        process.exit(1)
    }

    compiler.hooks.invalid.tap('compileInvalidate', () => {
        if (isInteractive) {
            clearConsole()
        }
        console.log('Compiling...')
        if (spawnedApp) {
            // TODO: handle fail?
            app.kill()
            spawnedApp = false
        }
    })

    let isFirstCompile = true

    compiler.hooks.done.tap('compileDone', stats => {
        if (isInteractive) {
            clearConsole()
        }

        const messages = formatWebpackMessages(stats.toJson({}, true))
        const isSuccessful =
            !messages.errors.length && !messages.warnings.length

        if (isSuccessful) {
            console.log(chalk.green('Compiled successfully!'))
        }

        isFirstCompile = false

        if (messages.errors.length) {
            const collapsedErrorLength = messages.errors.length - 1
            // Only keep the first error. Others are often indicative
            // of the same problem, but confuse the reader with noise.
            if (messages.errors.length > 1) {
                messages.errors.length = 1
            }
            console.log(chalk.red('Failed to compile.\n'))
            console.log(messages.errors.join('\n\n'))
            if (collapsedErrorLength) {
                console.log()
                console.log(
                    chalk.red(
                        collapsedErrorLength +
                            (collapsedErrorLength > 1
                                ? ' errors have'
                                : ' error has') +
                            ' been collapsed.'
                    )
                )
            }
            return
        }

        // Show warnings if no errors were found.
        if (messages.warnings.length) {
            console.log(chalk.yellow('Compiled with warnings.\n'))
            console.log(messages.warnings.join('\n\n'))

            console.log(
                '\nSearch for the ' +
                    chalk.cyan('keywords') +
                    ' to learn more about each warning.'
            )
            console.log(
                'To ignore, add ' +
                    chalk.yellow('// tslint:disable-next-line') +
                    ' to the line before.\n'
            )
        }

        // Start the app if everything there are no errors
        if (!spawnedApp && !messages.errors.length) {
            app = spawn('node', [
                path.join(config.output.path, config.output.filename),
            ])

            if (app.pid > 0) {
                spawnedApp = true
                console.log(chalk.green('App started!'))
            }

            app.on('error', e => {
                console.log(chalk.red('Failed to start app: ' + e.message))
                spawnedApp = false
            })
            app.stdout.pipe(process.stdout)
            app.stderr.pipe(process.stderr)

            app.on('exit', code => {
                console.log(
                    code > 0
                        ? chalk.red('App exited with code ' + code + '.')
                        : chalk.green('App exited with code ' + code + '.')
                )
            })
        }
    })

    return compiler
}

module.exports = {
    createCompiler,
}
